"""
fordev.generators.cnh
---------------------
"""

from fordev.core import fordev_request

from fordev.filters import data_format


def cnh(data_only: bool=True) -> str:
    """Geração aleatória de CNH (Carteira Nacional de Habilitação)."""

    resp = fordev_request(
        content_length=14,
        referer='gerador_de_cnh',
        payload={'acao': 'gerar_cnh'}
    )

    return data_format(data_only=data_only, data_dict=resp)



"""
fordev.validators.rg
--------------------
"""

from fordev.core import fordev_request

from fordev.filters import data_format

from fordev.validators.utils import _data_verification_and_normalize


def is_valid_rg(rg_code: str, data_only: bool=True) -> bool:
    """Verifique se o código do RG é válido.

    Parameters
    ----------
    rg_code
        O código do RG para verificação.
    """

    content_length = 35
    referer = 'validador_rg'
    payload = {
        'acao': 'validar_rg',
        'txt_rg': rg_code
    }

    resp = _data_verification_and_normalize(
        fordev_request(content_length, referer, payload)
    )

    return data_format(data_only=data_only, data_dict=resp)



"""
fordev.validators.voter_title
-----------------------------
"""

from fordev.core import fordev_request

from fordev.filters import data_format


def is_valid_voter_title(voter_title_code: str, data_only: bool=True) -> bool:
    """Verifique se o código do título de eleitor é válido.

    Parameters
    ----------
    voter_title_code
        O código do título de eleitor para verificação.
    """

    content_length = 59
    referer = 'validador_titulo_de_eleitor'
    payload = {
        'acao': 'validar_titulo_eleitor',
        'txt_titulo_eleitor': voter_title_code
    }

    resp = fordev_request(content_length, referer, payload)

    if resp.get('data', False):
        is_valid = resp['data'].split(' - ')[-2].lower() == 'verdadeiro'
        resp['data'] = is_valid

    return data_format(data_only=data_only, data_dict=resp)



"""
fordev.generators.bank_account
------------------------------
"""

from fordev.core import fordev_request

from fordev.filters import data_format
from fordev.filters import filter_bank_account_info

from fordev.validators.utils import raise_for_invalid_uf


def bank_account(bank: int=0, uf_code: str='', data_only: bool=True) -> dict:
    """Gere dados de conta bancária.

    Parameters
    ----------
    bank
        Recebe um valor númerico de 0 a 5 que representa a
        bandeira do banco da conta bancária a ser gerada.

        Consulte a doc para verificar as opções suportadas:
        https://fordev.rtfd.io/pt_BR/latest/fordev/generators.html
    """

    if not (0 <= bank <= 5):
        msg_error = (
            f'The bank code value "{bank}" is invalid. Enter a valid bank code.'
            ' The range is 0 to 5.'
        )

        raise ValueError(msg_error)

    bank = ['', 2, 121, 85, 120, 151][bank]

    uf_code = uf_code.upper()

    raise_for_invalid_uf(uf=uf_code, include_blank=True)

    resp = fordev_request(
        content_length=45,
        referer='gerador_conta_bancaria',
        payload={
            'acao': 'gerar_conta_bancaria',
            'estado': uf_code,
            'banco': bank
        }
    )

    resp['data'] = filter_bank_account_info(resp['data'])

    return data_format(data_only=data_only, data_dict=resp)



"""
fordev.validators
-----------------

Este módulo válida os dados utilizando o site `4Devs <https://www.4devs.com.br/>`_
e disponibiliza uma API simples para uso.

Use a função ``help()`` para mais informações:

>>> from fordev import validators
>>> help(validators)
Help on module fordev.validators in fordev:

NAME
    fordev.validators

DESCRIPTION
(...)

Ou consulte a documentação oficial.

Note
----
Muitas funções do módulo ``fordev.validators`` contém parâmetros em comum,
todos estão descritos abaixo.

More details in next section.

Parameter
---------
data_only: bool
    Se receber o valor ``True``, retorna somente os dados em texto puro.
    Se receber o valor ``False``, retorna um dicionário contendo uma chave ``msg`` e ``data`` ou ``error``
    contendo valores correspondentes à nomenclatura de suas chaves.

Sendo assim, sempre que o encontrar, utilize conforme o descrito acima.
"""

__all__ = [
    'is_valid_credit_card',
    'is_valid_bank_account',
    'is_valid_certificate',
    'is_valid_cnh',
    'is_valid_cnpj',
    'is_valid_cpf',
    'is_valid_pis_pasep',
    'is_valid_renavam',
    'is_valid_rg',
    'is_valid_voter_title',
    'is_valid_state_registration'
]

from fordev.validators.credit_card import is_valid_credit_card
from fordev.validators.bank_account import is_valid_bank_account
from fordev.validators.certificate import is_valid_certificate
from fordev.validators.cnh import is_valid_cnh
from fordev.validators.cnpj import is_valid_cnpj
from fordev.validators.cpf import is_valid_cpf
from fordev.validators.pis_pasep import is_valid_pis_pasep
from fordev.validators.renavam import is_valid_renavam
from fordev.validators.rg import is_valid_rg
from fordev.validators.voter_title import is_valid_voter_title
from fordev.validators.state_registration import is_valid_state_registration



"""
fordev.generators.rg
--------------------
"""

from fordev.core import fordev_request

from fordev.filters import data_format


def rg(formatting: bool=True, data_only: bool=True) -> str:
    """Gere o código do RG(Registro Geral) aleatório, emitido por SSP-SP."""

    resp = fordev_request(
        content_length=25,
        referer='gerador_de_rg',
        payload={
            'acao': 'gerar_rg',
            'pontuacao': 'S' if formatting else 'N',
        }
    )

    return data_format(data_only=data_only, data_dict=resp)



"""
fordev.generators.cpf
---------------------
"""

from fordev.core import fordev_request

from fordev.filters import data_format

from fordev.validators.utils import raise_for_invalid_uf


def cpf(uf_code: str='', formatting: bool=True, data_only: bool=True) -> str:
    """Gere o código de um CPF(Cadastro de Pessoas Físicas) aleatório."""

    uf_code = uf_code.upper()

    raise_for_invalid_uf(uf=uf_code, include_blank=True)

    resp = fordev_request(
        content_length=38 if uf_code == '' else 40,
        referer='gerador_de_cpf',
        payload={
            'acao': 'gerar_cpf',
            'pontuacao': 'S' if formatting else 'N',
            'cpf_estado': uf_code
        }
    )

    return data_format(data_only=data_only, data_dict=resp)



"""
fordev.validators.bank_account
------------------------------
"""

from fordev.core import fordev_request

from fordev.filters import data_format

from fordev.validators.utils import _data_verification_and_normalize


def is_valid_bank_account(bank: int, agency: str, account: str, data_only: bool=True) -> bool:
    """Verifique se os dados da conta bancária são válidos.

    Parameters
    ----------
    bank
        A bandeira do banco da conta bancária que deseja validar os dados.

        Consulte a doc para verificar as opções suportadas:
        https://fordev.rtfd.io/pt_BR/latest/fordev/generators.html

    agency
        O código da agência bancária para verificação.

    account
        O código da conta bancária para verificação.
   """

    if not (1 <= bank <= 5):
        msg_error = (
            f'The bank code value "{bank}" is invalid. Enter a valid bank code.'
            ' The range is 1 to 5.'
        )

        raise ValueError(msg_error)

    bank = [2, 121, 85, 120, 151][bank - 1]

    content_length = 66
    referer = 'validador_conta_bancaria'
    payload = {
        'acao': 'validar_conta_bancaria',
        'banco': bank,
        'agencia': agency,
        'conta': account
    }

    resp = _data_verification_and_normalize(
        fordev_request(content_length, referer, payload)
    )

    return data_format(data_only=data_only, data_dict=resp)



"""
fordev.consts
-------------

Este módulo contém todas as variáveis com valores constantes usadas no pacote fordev.
"""

URL_4DEV_API = 'https://www.4devs.com.br/ferramentas_online.php'

ALL_UF_CODE = [
    'AC', 'AL', 'AP', 'AM', 'BA', 'CE', 'ES', 'GO', 'MA',
    'MT', 'MS', 'MG','PA', 'PB', 'PR', 'PE', 'PI', 'RJ',
    'RN', 'RS', 'RO', 'RR', 'SC', 'SP', 'SE', 'TO', 'DF'
]

ALL_VEHICLE_BRANDS = {
    1: {'brand_name': 'Acura', 'code': 1},
    2: {'brand_name': 'Agrale', 'code': 2},
    3: {'brand_name': 'Alfa Romeo', 'code': 3},
    4: {'brand_name': 'AM Gen', 'code': 4},
    5: {'brand_name': 'Asia Motors', 'code': 5},
    6: {'brand_name': 'ASTON MARTIN', 'code': 189},
    7: {'brand_name': 'Audi', 'code': 6},
    8: {'brand_name': 'BMW', 'code': 7},
    9: {'brand_name': 'BRM', 'code': 8},
    10: {'brand_name': 'Buggy', 'code': 9},
    11: {'brand_name': 'Bugre', 'code': 123},
    12: {'brand_name': 'Cadillac', 'code': 10},
    13: {'brand_name': 'CBT Jipe', 'code': 11},
    14: {'brand_name': 'CHANA', 'code': 136},
    15: {'brand_name': 'CHANGAN', 'code': 182},
    16: {'brand_name': 'CHERY', 'code': 161},
    17: {'brand_name': 'Chrysler', 'code': 12},
    18: {'brand_name': 'Citroen', 'code': 13},
    19: {'brand_name': 'Cross Lander', 'code': 14},
    20: {'brand_name': 'Daewoo', 'code': 15},
    21: {'brand_name': 'Daihatsu', 'code': 16},
    22: {'brand_name': 'Dodge', 'code': 17},
    23: {'brand_name': 'EFFA', 'code': 147},
    24: {'brand_name': 'Engesa', 'code': 18},
    25: {'brand_name': 'Envemo', 'code': 19},
    26: {'brand_name': 'Ferrari', 'code': 20},
    27: {'brand_name': 'Fiat', 'code': 21},
    28: {'brand_name': 'Fibravan', 'code': 149},
    29: {'brand_name': 'Ford', 'code': 22},
    30: {'brand_name': 'FOTON', 'code': 190},
    31: {'brand_name': 'Fyber', 'code': 170},
    32: {'brand_name': 'GEELY', 'code': 199},
    33: {'brand_name': 'GM - Chevrolet', 'code': 23},
    34: {'brand_name': 'GREAT WALL', 'code': 153},
    35: {'brand_name': 'Gurgel', 'code': 24},
    36: {'brand_name': 'HAFEI', 'code': 152},
    37: {'brand_name': 'Honda', 'code': 25},
    38: {'brand_name': 'Hyundai', 'code': 26},
    39: {'brand_name': 'Isuzu', 'code': 27},
    40: {'brand_name': 'JAC', 'code': 177},
    41: {'brand_name': 'Jaguar', 'code': 28},
    42: {'brand_name': 'Jeep', 'code': 29},
    43: {'brand_name': 'JINBEI', 'code': 154},
    44: {'brand_name': 'JPX', 'code': 30},
    45: {'brand_name': 'Kia Motors', 'code': 31},
    46: {'brand_name': 'Lada', 'code': 32},
    47: {'brand_name': 'LAMBORGHINI', 'code': 171},
    48: {'brand_name': 'Land Rover', 'code': 33},
    49: {'brand_name': 'Lexus', 'code': 34},
    50: {'brand_name': 'LIFAN', 'code': 168},
    51: {'brand_name': 'LOBINI', 'code': 127},
    52: {'brand_name': 'Lotus', 'code': 35},
    53: {'brand_name': 'Mahindra', 'code': 140},
    54: {'brand_name': 'Maserati', 'code': 36},
    55: {'brand_name': 'Matra', 'code': 37},
    56: {'brand_name': 'Mazda', 'code': 38},
    57: {'brand_name': 'Mercedes-Benz', 'code': 39},
    58: {'brand_name': 'Mercury', 'code': 40},
    59: {'brand_name': 'MG', 'code': 167},
    60: {'brand_name': 'MINI', 'code': 156},
    61: {'brand_name': 'Mitsubishi', 'code': 41},
    62: {'brand_name': 'Miura', 'code': 42},
    63: {'brand_name': 'Nissan', 'code': 43},
    64: {'brand_name': 'Peugeot', 'code': 44},
    65: {'brand_name': 'Plymouth', 'code': 45},
    66: {'brand_name': 'Pontiac', 'code': 46},
    67: {'brand_name': 'Porsche', 'code': 47},
    68: {'brand_name': 'RAM', 'code': 185},
    69: {'brand_name': 'RELY', 'code': 186},
    70: {'brand_name': 'Renault', 'code': 48},
    71: {'brand_name': 'Rolls-Royce', 'code': 195},
    72: {'brand_name': 'Rover', 'code': 49},
    73: {'brand_name': 'Saab', 'code': 50},
    74: {'brand_name': 'Saturn', 'code': 51},
    75: {'brand_name': 'Seat', 'code': 52},
    76: {'brand_name': 'SHINERAY', 'code': 183},
    77: {'brand_name': 'smart', 'code': 157},
    78: {'brand_name': 'SSANGYONG', 'code': 125},
    79: {'brand_name': 'Subaru', 'code': 54},
    80: {'brand_name': 'Suzuki', 'code': 55},
    81: {'brand_name': 'TAC', 'code': 165},
    82: {'brand_name': 'Toyota', 'code': 56},
    83: {'brand_name': 'Troller', 'code': 57},
    84: {'brand_name': 'Volvo', 'code': 58},
    85: {'brand_name': 'VW - VolksWagen', 'code': 59},
    86: {'brand_name': 'Wake', 'code': 163},
    87: {'brand_name': 'Walk', 'code': 120}
}

ALL_BANK_FLAGS = {
    1: 'master',
    2: 'visa16',
    3: 'amex',
    4: 'diners',
    5: 'discover',
    6: 'enroute',
    7: 'jcb',
    8: 'voyager',
    9: 'hiper',
    10: 'aura'
}

ALL_BANK_FLAGS_2 = {
    1: 'MasterCard',
    2: 'Visa',
    3: 'Visa Electron',
    4: 'American Express',
    5: 'Diners Club',
    6: 'Discover',
    7: 'Enroute',
    8: 'JCB',
    9: 'Maestro',
    10: 'Solo',
    11: 'Switch',
    12: 'LaserCard'
}



"""
fordev.validators.credit_card
-----------------------------
"""

from fordev.core import fordev_request

from fordev.consts import ALL_BANK_FLAGS_2

from fordev.filters import data_format

from fordev.validators.utils import _data_verification_and_normalize


def is_valid_credit_card(flag: int, credit_card_code: str, data_only: bool=True) -> bool:
    """Verifique se o código do cartão de crédito é válido.

    Parameters
    ----------
    flag
        A bandeira do cartão de crédito que deseja validar o código.

        Consulte a doc para verificar as opções suportadas:
        https://fordev.rtfd.io/pt_BR/latest/fordev/generators.html

    credit_card_code
        O código do cartão de crédito para verificação.
   """

    if not (1 <= flag <= 12):
        msg_error = (
            f'The flag credit card code value "{flag}" is invalid.'
            ' Enter a valid flag credit card code.'
            ' The range is 1 to 12.'
        )

        raise ValueError(msg_error)

    flag = ALL_BANK_FLAGS_2[flag]

    content_length = 68
    referer = 'validador_numero_cartao_credito'
    payload = {
        'acao': 'validar_cc',
        'txt_cc': credit_card_code,
        'bandeira': flag
    }

    resp = _data_verification_and_normalize(
        fordev_request(content_length, referer, payload)
    )

    return data_format(data_only=data_only, data_dict=resp)



"""
fordev.generators.voter_title
-----------------------------
"""

from fordev.core import fordev_request

from fordev.validators.utils import raise_for_invalid_uf

from fordev.filters import data_format


def voter_title(uf_code: str, data_only: bool=True) -> str:
    """Gere o código do título de eleitor aleatório, conforme o UF especificado."""

    uf_code = uf_code.upper()

    raise_for_invalid_uf(uf=uf_code)

    resp = fordev_request(
        content_length=35,
        referer='gerador_de_titulo_de_eleitor',
        payload={
            'acao': 'gerar_titulo_eleitor',
            'estado': uf_code
        }
    )

    return data_format(data_only=data_only, data_dict=resp)



"""
fordev.generators.city
----------------------
"""

from fordev.core import fordev_request

from fordev.validators.utils import raise_for_invalid_uf

from fordev.filters import filter_city_name
from fordev.filters import data_format


def city(uf_code: str='SP', data_only: bool=True) -> list:
    """Obtenha as cidades do UF especificado."""

    uf_code = uf_code.upper()

    raise_for_invalid_uf(uf=uf_code)

    resp = fordev_request(
        content_length=35,
        referer='gerador_de_pessoas',
        payload={
            'acao': 'carregar_cidades',
            'cep_estado': uf_code
        }
    )

    resp['data'] = filter_city_name(resp['data'])

    return data_format(data_only=data_only, data_dict=resp)



"""
fordev.generators.renavam
-------------------------
"""

from fordev.core import fordev_request

from fordev.filters import data_format


def renavam(data_only: bool=True) -> str:
    """Gere o código do RENAVAM(Registro Nacional de Veículos Automotores) aleatório."""

    resp = fordev_request(
        content_length=18,
        referer='gerador_de_renavam',
        payload={'acao': 'gerar_renavam'}
    )

    return data_format(data_only=data_only, data_dict=resp)



"""
fordev.validators.utils
-----------------------
"""

__all__ = [
    '_data_verification_and_normalize',
    'raise_for_invalid_uf'
]

from fordev.consts import ALL_UF_CODE


def _data_verification_and_normalize(data: dict) -> dict:
    """"Verifique se a key existe e se o valor é válido.
    Se válido, substítui para um novo formato.

    Parameters
    ----------
    data
        Um dicionário de dados para verificação e mudança de formato.
    """

    data = data.copy()

    if data.get('data', False):
        is_valid = data['data'].split(' - ')[-1].lower() == 'verdadeiro'
        data['data'] = is_valid

    return data


def raise_for_invalid_uf(uf: str, include_blank=False):
    """Levanta uma exceção if o código UF for inválido.

    Parameters
    ----------
    include_blank
        Algumas funções enviam um UF em branco, para considerá-lo
        defina ``include_blank`` como ``True``.
    """

    ufs = ALL_UF_CODE.copy()

    if include_blank:
        ufs.append('')

    if uf not in ufs:
        msg_error = (
            f'The UF code "{uf}" is invalid. Enter a valid UF code. Ex: SP, RJ, PB...'
            ' More info about UF in: https://pt.wikipedia.org/wiki/Subdivis%C3%B5es_do_Brasil'
        )

        raise ValueError(msg_error)



"""
fordev.generators.vehicle
-------------------------
"""

from random import sample as random_sample

from fordev.core import fordev_request

from fordev.consts import ALL_VEHICLE_BRANDS

from fordev.validators.utils import raise_for_invalid_uf

from fordev.filters import filter_vehicle_info
from fordev.filters import data_format


def vehicle(
        brand_code: int=0,
        uf_code: str='',
        formatting: bool=True,
        data_only: bool=True
    ) -> dict:
    """Gere dados de veículo aleatório.

    Parameters
    ----------
    brand
        Recebe um valor númerico de 0 a 87 que representa a marca do carro para
        geração dos dados aleatórios.

        Consulte a doc para verificar as opções suportadas:
        https://fordev.rtfd.io/pt_BR/latest/fordev/generators.html
    """

    if not (0 <= brand_code <= 87):
        msg_error = (
            f'The vehicle brand code value "{brand_code}" is invalid.'
            ' Enter a valid vehicle brand code.'
            ' The range is 0 to 87.'
        )

        raise ValueError(msg_error)

    if brand_code != 0:
        brand_code = ALL_VEHICLE_BRANDS[brand_code]['code']
    else:
        brand_code = ''

    uf_code = uf_code.upper()

    raise_for_invalid_uf(uf=uf_code, include_blank=True)

    resp = fordev_request(
        content_length=62,
        referer='gerador_de_veiculos',
        payload={
            'acao': 'gerar_veiculo',
            'pontuacao': 'S' if formatting else 'N',
            'estado': uf_code,
            'fipe_codigo_marca': brand_code
        }
    )

    resp['data'] = filter_vehicle_info(resp['data'])

    return data_format(data_only=data_only, data_dict=resp)


def vehicle_brand(n: int=1, data_only: bool=True) -> list:
    """Obtenha o nome de marca(s) de veículo(s).

    Parameters
    ----------
    n
        Recebe o número de marcas de veículos a ser gerado. O valor mínimo é 1 e o máximo é 87.
    """

    if not (1 <= n <= 87):
        msg_error = (
            f'The n value "{n}" is invalid. Enter a valid number of UF.'
            ' The range is 1 to 27 UF code.'
        )

        raise ValueError(msg_error)

    full_data = {
        'msg': 'success',
        'data': random_sample(
            [v_brand['brand_name'] for v_brand in ALL_VEHICLE_BRANDS.values()],
            n
        )
    }

    if data_only:
        return full_data['data']
    else:
        return full_data


def vehicle_plate(uf_code: str='', formatting: bool=True, data_only: bool=True) -> str:
    """Gere o código da placa de veículo aleatório."""

    uf_code = uf_code.upper()

    raise_for_invalid_uf(uf=uf_code, include_blank=True)

    resp = fordev_request(
        content_length=36 if uf_code == '' else 38,
        referer='gerador_de_placa_automoveis',
        payload={
            'acao': 'gerar_placa',
            'pontuacao': 'S' if formatting else 'N',
            'estado':uf_code
        }
    )

    return data_format(data_only=data_only, data_dict=resp)



"""
fordev.generators
-----------------

Este módulo coleta dados aleatórios gerados pelo site `4Devs <https://www.4devs.com.br/>`_
e disponíbiliza uma API simples para uso.

Use a função ``help()`` para mais informações:

>>> from fordev import generators
>>> help(generators)
Help on module fordev.generators in fordev:

NAME
    fordev.generators

DESCRIPTION
(...)

Ou consulte a documentação oficial.

Note
----
Muitas funções do módulo ``fordev.generators`` contém parâmetros em comum,
todos estão descrito abaixo.

Parameters
----------
uf_code: str
    Recebe o Código da **Unidade Federativa** para geração do dado.

    Caso não saiba o que é ou não conheça o do estado que necessita,
    obtenha mais informações em: https://pt.wikipedia.org/wiki/Subdivis%C3%B5es_do_Brasil

formatting: bool
    Se receber o valor ``True``, retorna os dados formatados.
    Se receber o valor ``False``, retorna os dados sem formatação.

data_only: bool
    Se receber o valor ``True``, retorna somente os dados em texto puro.
    Se receber o valor ``False``, retorna um dicionário contendo uma chave ``msg`` e ``data`` ou ``error``
    contendo valores correspondentes à nomenclatura de suas chaves.

Sendo assim, sempre que os encontrar, utilize conforme o descrito acima.
"""

__all__ = [
    'certificate',
    'cnh',
    'bank_account',
    'cpf',
    'pis_pasep',
    'renavam',
    'vehicle',
    'vehicle_brand',
    'vehicle_plate',
    'cnpj',
    'rg',
    'state_registration',
    'voter_title',
    'credit_card',
    'people',
    'company',
    'uf',
    'city'
]

from fordev.generators.certificate import certificate
from fordev.generators.cnh import cnh
from fordev.generators.bank_account import bank_account
from fordev.generators.cpf import cpf
from fordev.generators.pis_pasep import pis_pasep
from fordev.generators.renavam import renavam
from fordev.generators.vehicle import vehicle, vehicle_brand, vehicle_plate
from fordev.generators.cnpj import cnpj
from fordev.generators.rg import rg
from fordev.generators.state_registration import state_registration
from fordev.generators.voter_title import voter_title
from fordev.generators.credit_card import credit_card
from fordev.generators.people import people
from fordev.generators.company import company
from fordev.generators.uf import uf
from fordev.generators.city import city



"""
fordev.generators.company
-------------------------
"""

from fordev.core import fordev_request

from fordev.validators.utils import raise_for_invalid_uf

from fordev.filters import filter_company_info
from fordev.filters import data_format


def company(uf_code: str='SP', age: int=1, formatting: bool=True, data_only: bool=True) -> dict:
    """Gere dados de companhia (empresa/organização) aleatório.

    Parameters
    ----------
    age
        Representa o tempo de existência da companhia (a idade da companhia).
    """

    uf_code = uf_code.upper()

    raise_for_invalid_uf(uf=uf_code)

    if not (1 <= age <= 30):
        msg_error = (
            f'The company age value "{age}" is invalid. Enter a valid company age.'
            ' The range is 1 to 30.'
        )

        raise ValueError(msg_error)

    resp = fordev_request(
        content_length=48,
        referer='gerador_de_empresas',
        payload={
            'acao': 'gerar_empresa',
            'pontuacao': 'S' if formatting else 'N',
            'estado': uf_code,
            'idade': age
        }
    )

    resp['data'] = filter_company_info(resp['data'])

    return data_format(data_only=data_only, data_dict=resp)



"""
fordev.core
-----------

Este módulo é o core para criar e manipular requests para a API do site 4devs.
"""

__all__ = ['fordev_request']

import requests

from fordev.consts import URL_4DEV_API


def _create_headers(content_length: int, referer: str) -> dict:
    """Gere o header para ser enviado em requests HTTP para o site 4devs.

    Parameters
    ----------
    content_length
        Indica o tamanho do entity-body, em bytes, enviados no header para o destinatário.

    referer
        Referência a ação a ser executada pela API do site 4devs.
        Pode-se interpretar como o endpoint do serviço a ser disponibilizado.

    """

    headers = {
        'user-agent': (
            'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko)'
            'Chrome/106.0.0.0 Safari/537.36'
        ),
        'authority': 'www.4devs.com.br',
        'method': 'POST',
        'path': '/ferramentas_online.php',
        'scheme': 'https',
        'accept': '*/*',
        'accept-encoding': 'gzip, deflate, br',
        'accept-language': 'pt,en-US;q=0.9,en;q=0.8',
        'content-length': str(content_length),
        'content-type': 'application/x-www-form-urlencoded; charset=UTF-8',
        'dnt': '1',
        'origin': 'https://www.4devs.com.br',
        'referer': f'https://www.4devs.com.br/{referer}',
        'sec-fetch-dest': 'empty',
        'sec-fetch-mode': 'cors',
        'sec-fetch-site': 'same-origin',
        'x-requested-with': 'XMLHttpRequest'
    }

    return headers


def fordev_request(content_length: int, referer: str, payload: dict) -> dict:
    """Cria uma request HTTP a API do site 4devs e
    retorna seu conteúdo em formato de dicionário.

    Parameters
    ----------
    content_length
        Indica o tamanho do entity-body, em bytes, enviados no header para o destinatário.

    referer
        Referência a ação a ser executada pela API do site 4devs.
        Pode-se interpretar como o endpoint do serviço a ser disponibilizado.

    payload
        Um dicionário de dados contendo a ação e outros dados solicitados pela API.
    """

    try:
        response = requests.post(
            url=URL_4DEV_API,
            headers=_create_headers(content_length, referer),
            data=payload,
            timeout=10
        )

        response.raise_for_status()

        return {
            'msg': 'success',
            'data': response.text
        }

    except (requests.RequestException, requests.HTTPError) as err:
        return {
            'msg': 'failed',
            'error': err
        }



"""
fordev.generators.people
------------------------
"""

from json import loads as json_loads

from fordev.core import fordev_request

from fordev.validators.utils import raise_for_invalid_uf


def people(
        n: int=1,
        sex: str='R',
        age: int=0,
        uf_code: str='',
        formatting: bool=True,
        data_only: bool=True
    ) -> str:
    """Gere dados de pessoa(s) aleatório(s)

    Parameters
    ----------
    n
        O número de pessoas a ter dados gerados. O mínimo é 1 e o máximo é 30.

    sex
        Uma string representando o sexo da pessoa para geração dos dados.

        Consulte a doc para verificar as opções suportadas:
        https://fordev.rtfd.io/pt_BR/latest/fordev/generators.html

    age
        A idade da pessoa para geração dos dados. A idade mínima é 18 e a máxima é 80.
    """

    sex = sex.upper()

    uf_code = uf_code.upper()

    if not (1 <= n <= 30):
        msg_error = (
            f'The n value "{n}" is invalid. Enter a valid number of people.'
            ' The range is 1 to 30 peoples.'
        )

        raise ValueError(msg_error)

    if sex not in ['M', 'F', 'R']:
        msg_error = (
            f'The sex "{sex}" is invalid. Enter a valid sex.'
            ' Ex: "M" = Male, "F" = Feminine or "R" = Random.'
        )

        raise ValueError(msg_error)

    if not (18 <= age <= 80) and age != 0:
        msg_error = (
            f'The age "{age}" is invalid. Enter a valid age.'
            ' The range is 18 to 80 age.'
        )

        raise ValueError(msg_error)

    raise_for_invalid_uf(uf=uf_code, include_blank=True)

    resp = fordev_request(
        content_length=99,
        referer='gerador_de_pessoas',
        payload={
            'acao': 'gerar_pessoa',
            'sexo': 'H' if sex == 'M' else 'M' if sex == 'F' else 'I',
            'pontuacao': 'S' if formatting else 'N',
            'idade': age,
            'cep_estado': uf_code,
            'txt_qtde': n,
            'cep_cidade': 'Selecione o estado!' if uf_code == '' else ''
        }
    )

    if data_only and resp['msg'] == 'success':
        return json_loads(resp['data'])

    if resp['msg'] == 'success':

        resp['data'] = json_loads(resp['data'])

        return resp

    return resp



"""
fordev.validators.certificate
-----------------------------
"""

from fordev.core import fordev_request

from fordev.filters import data_format

from fordev.validators.utils import _data_verification_and_normalize


def is_valid_certificate(certificate_code: str, data_only: bool=True) -> bool:
    """Verifique se o código da Certidão (birth, wedding, religious wedding and death) é válido.

    Parameters
    ----------
    certificate_code
        O código da certidão para verificação.
    """

    content_length = 75
    referer = 'validador_certidoes'
    payload = {
        'acao': 'validar_certidao',
        'txt_certidao': certificate_code
    }

    resp = _data_verification_and_normalize(
        fordev_request(content_length, referer, payload)
    )

    return data_format(data_only=data_only, data_dict=resp)



"""
fordev.filters
--------------

Este módulo é usado para filtrar os dados contidos na estrutura HTML
retornada pela API do site 4devs e converter para uma estrutura de
dicionário Python.
"""

__all__ = [
    'data_format',
    'filter_bank_account_info',
    'filter_vehicle_info',
    'filter_credit_card_info',
    'filter_company_info',
    'filter_city_name'
]

from bs4 import BeautifulSoup


def data_format(data_only: bool, data_dict: dict) -> dict:
    """Filtra os dados conforme especificado."""

    if data_only and data_dict['msg'] == 'success':
        return data_dict['data']

    return data_dict


def filter_bank_account_info(html: str) -> dict:
    """Filtra dados de conta bancária contidos na estrutura HTML.

    Parameters
    ----------
    html
        Uma estrutra HTML contendo dados de conta bancária
        retornados pela API do site 4devs.
    """

    soup = BeautifulSoup(html, 'html.parser')

    labels = [div.text for div in soup.find_all('div', 'output-subtitle')]

    data_of_bank_account = [div.text for div in soup.find_all('div', 'output-txt')]

    data_dict = dict(
        zip(labels, data_of_bank_account)
    )

    return data_dict


def filter_vehicle_info(html: str) -> dict:
    """Filtra dados de veículo contido na estrutura HTML.

    Parameters
    ----------
    html
        Uma estrutra HTML contendo dados de veículo
        retornados pela API do site 4devs.
    """

    soup = BeautifulSoup(html, 'html.parser')

    labels = [div.text[:-1] for div in soup.find_all('strong')]

    data_of_vehicle = [input_.get('value') for input_ in soup.find_all('input', 'margem_menor')]

    data_dict = dict(
        zip(labels, data_of_vehicle)
    )

    return data_dict


def filter_credit_card_info(html: str) -> dict:
    """Filtra dados de cartão de crédito contidos na estrutura HTML.

    Parameters
    ----------
    html
        Uma estrutra HTML contendo dados de cartão de crédito
        retornados pela API do site 4devs.
    """

    soup = BeautifulSoup(html, 'html.parser')

    labels = [div.text for div in soup.find_all('div', 'output-subtitle')]

    data_of_credit_card = [div.text.strip() for div in soup.find_all('div', 'output-txt')]

    data_dict = dict(
        zip(labels, data_of_credit_card)
    )

    return data_dict


def filter_company_info(html: str) -> dict:
    """Filtra dados de companhia (empresa/organização) contidos na estrutura HTML.

    Parameters
    ----------
    html
        Uma estrutra HTML contendo dados de companhia (empresa/organização)
        retornados pela API do site 4devs.
    """

    soup = BeautifulSoup(html, 'html.parser')

    labels = [div.text[:-1] for div in soup.find_all('strong')]

    data_of_company = [input_.get('value') for input_ in soup.find_all('input', 'margem_menor')]

    data_dict = dict(
        zip(labels, data_of_company)
    )

    return data_dict


def filter_city_name(html: str) -> list:
    """Filtra dados de cidade contidos na estrutura HTML.

    Parameters
    ----------
    html
        Uma estrutra HTML contendo dados de uma cidade
        retornados pela API do site 4devs.
    """

    soup = BeautifulSoup(html, 'html.parser')

    return [
        option.text for option in soup.find_all('option')[1:]
    ]



"""
fordev.generators.uf
--------------------
"""

from random import sample as random_sample

from fordev.consts import ALL_UF_CODE


def uf(n: int=1, data_only: bool=True) -> list:
    """Gere o código da UF(Unidade Federativa) aleatório.

    Parameters
    ----------
    n
        O número de UF's para geração do dado. O mínimo é 1 e o máximo é 27.
    """

    if not (1 <= n <= 27):
        msg_error = (
            f'The n value "{n}" is invalid. Enter a valid number of UF.'
            ' The range is 1 to 27 UF code.'
        )

        raise ValueError(msg_error)

    full_data = {
        'msg': 'success',
        'data': random_sample(ALL_UF_CODE, n)
        }

    if data_only:
        return full_data['data']
    else:
        return full_data



"""
fordev.validators.state_registration
------------------------------------
"""

from fordev.core import fordev_request

from fordev.filters import data_format

from fordev.validators.utils import _data_verification_and_normalize
from fordev.validators.utils import raise_for_invalid_uf


def is_valid_state_registration(
        uf_code: str,
        state_registration_code: str,
        data_only: bool=True
    ) -> bool:
    """Verifique se o código do registro estadual é válido.

    Parameters
    ----------
    uf_code
        O código UF(Unidade Federativa) do estado que pertence o registro estadual.

        Mais informações: https://pt.wikipedia.org/wiki/Subdivis%C3%B5es_do_Brasil

    state_registration_code
        O código do registro estadual para verificação.
    """

    uf_code = uf_code.upper()

    raise_for_invalid_uf(uf=uf_code)

    resp = _data_verification_and_normalize(
        fordev_request(
            content_length=48,
            referer='validar_inscricao_estadual',
            payload={
                'acao': 'validar_ie',
                'txt_ie': state_registration_code,
                'estado': uf_code
            }
        )
    )

    return data_format(data_only=data_only, data_dict=resp)



"""
fordev.generators.certificate
-----------------------------
"""

from fordev.core import fordev_request

from fordev.filters import data_format


def certificate(type_: str='I', formatting: bool=True, data_only: bool=True) -> str:
    """Gere o código de certidões (birth, wedding, religious wedding and death) aleatórias.

    Parameters
    ----------
    type_
        O tipo da certidão para geração do código.

        Consulte a doc para verificar as opções suportadas:
        https://fordev.rtfd.io/pt_BR/latest/fordev/generators.html
    """

    type_ = type_.upper()

    certificate_types = {
        'I': 'Indiferente',
        'B': 'nascimento',
        'W': 'casamento',
        'R': 'casamento_religioso',
        'D': 'obito'
    }

    if not certificate_types.get(type_, False):
        msg_error = (
            f'The certificate type "{type_}" is invalid. Enter a valid type.'
            ' Ex: "B" = Birth, "W" = Wedding, "R" = Religious Wedding,'
            '"D" = Death and "I" = Indifferent (Default).'
        )
        raise ValueError(msg_error)

    resp = fordev_request(
        content_length=67,
        referer='gerador_numero_certidoes',
        payload={
            'acao': 'gerador_certidao',
            'pontuacao': 'S' if formatting else 'N',
            'tipo_certidao': certificate_types.get(type_)
        }
    )

    return data_format(data_only=data_only, data_dict=resp)



"""
fordev.generators.cpnj
----------------------
"""

from fordev.core import fordev_request

from fordev.filters import data_format


def cnpj(formatting: bool=True, data_only: bool=True) -> str:
    """Gere o código do CNPJ(Cadastro Nacional da Pessoa Jurídica) aleatório."""

    resp = fordev_request(
        content_length=27,
        referer='gerador_de_cnpj',
        payload={
            'acao': 'gerar_cnpj',
            'pontuacao': 'S' if formatting else 'N',
        }
    )

    return data_format(data_only=data_only, data_dict=resp)



"""
fordev.generators.credit_card
-----------------------------
"""

from random import choice as random_choice

from fordev.core import fordev_request

from fordev.consts import ALL_BANK_FLAGS

from fordev.filters import filter_credit_card_info
from fordev.filters import data_format


def credit_card(bank: int=0, formatting: bool=True, data_only: bool=True) -> dict:
    """Gere dados de cartão de crédito aleatório.

    Parameters
    ----------
    bank
        Recebe um valor númerico de 0 a 10 representando a
        bandeira do cartão de crédito a ser gerado.

        Consulte a doc para verificar as opções suportadas:
        https://fordev.rtfd.io/pt_BR/latest/fordev/generators.html
    """

    if not (0 <= bank <= 10):
        msg_error = (
            f'The bank code value "{bank}" is invalid. Enter a valid bank code.'
            ' The range is 0 to 10.'
        )

        raise ValueError(msg_error)

    if bank != 0:
        bank = ALL_BANK_FLAGS[bank]
    else:
        bank = random_choice(
            list(ALL_BANK_FLAGS.values())
        )

    resp = fordev_request(
        content_length=43,
        referer='gerador_de_numero_cartao_credito',
        payload={
            'acao': 'gerar_cc',
            'pontuacao': 'S' if formatting else 'N',
            'bandeira': bank
        }
    )

    resp['data'] = filter_credit_card_info(resp['data'])

    return data_format(data_only=data_only, data_dict=resp)



"""
fordev.generators.pis_pasep
---------------------------
"""

from fordev.core import fordev_request

from fordev.filters import data_format


def pis_pasep(formatting: bool=True, data_only: bool=True) -> str:
    """Gere o código do PIS/PASEP aleatório."""

    resp = fordev_request(
        content_length=26,
        referer='gerador_de_pis_pasep',
        payload={
            'acao': 'gerar_pis',
            'pontuacao': 'S' if formatting else 'N'
        }
    )

    return data_format(data_only=data_only, data_dict=resp)
